{% extends "basic/search.html" %}

{% block scriptwarning %}
<div id="fallback" class="alert alert-warning">
<script>$('#fallback').hide();</script>
<h5>{{ _('Warning') }}</h5>
<p>
{% trans %}Please activate JavaScript to enable the search
functionality.{% endtrans %}
</p>
</div>
{% endblock %}

{% block searchbox %}
{% endblock %}
