// Copyright (c) 2020-2024 Advanced Micro Devices, Inc. All rights reserved.
// Headers from LSDp CL1744173
#ifndef _vpe_6_1_0_DEFAULT_HEADER
#define _vpe_6_1_0_DEFAULT_HEADER


// addressBlock: vpe_vpedec
#define regVPEC_DEC_START_DEFAULT                                                 0x00000000
#define regVPEC_UCODE_ADDR_DEFAULT                                                0x00000000
#define regVPEC_UCODE_DATA_DEFAULT                                                0x00000000
#define regVPEC_F32_CNTL_DEFAULT                                                  0x08084001
#define regVPEC_MMHUB_CNTL_DEFAULT                                                0x00000000
#define regVPEC_MMHUB_TRUSTLVL_DEFAULT                                            0x77777777
#define regVPEC_VPEP_CTRL_DEFAULT                                                 0x00000002
#define regVPEC_CLK_CTRL_DEFAULT                                                  0x00000000
#define regVPEC_PG_CNTL_DEFAULT                                                   0x00000015
#define regVPEC_POWER_CNTL_DEFAULT                                                0x00000000
#define regVPEC_CNTL_DEFAULT                                                      0x0150c401
#define regVPEC_CNTL1_DEFAULT                                                     0x05001000
#define regVPEC_CNTL2_DEFAULT                                                     0x400400c9
#define regVPEC_GB_ADDR_CONFIG_DEFAULT                                            0x00000242
#define regVPEC_GB_ADDR_CONFIG_READ_DEFAULT                                       0x00000242
#define regVPEC_PROCESS_QUANTUM0_DEFAULT                                          0x00000000
#define regVPEC_PROCESS_QUANTUM1_DEFAULT                                          0x00000000
#define regVPEC_CONTEXT_SWITCH_THRESHOLD_DEFAULT                                  0x0000006b
#define regVPEC_GLOBAL_QUANTUM_DEFAULT                                            0x00000000
#define regVPEC_WATCHDOG_CNTL_DEFAULT                                             0x00000000
#define regVPEC_ATOMIC_CNTL_DEFAULT                                               0x00000200
#define regVPEC_UCODE_VERSION_DEFAULT                                             0x00000000
#define regVPEC_MEMREQ_BURST_CNTL_DEFAULT                                         0x000006ff
#define regVPEC_TIMESTAMP_CNTL_DEFAULT                                            0x00000000
#define regVPEC_GLOBAL_TIMESTAMP_LO_DEFAULT                                       0x00000000
#define regVPEC_GLOBAL_TIMESTAMP_HI_DEFAULT                                       0x00000000
#define regVPEC_FREEZE_DEFAULT                                                    0x00000000
#define regVPEC_CE_CTRL_DEFAULT                                                   0x00000000
#define regVPEC_RELAX_ORDERING_LUT_DEFAULT                                        0xc0000806
#define regVPEC_CREDIT_CNTL_DEFAULT                                               0x00014840
#define regVPEC_SCRATCH_RAM_DATA_DEFAULT                                          0x00000000
#define regVPEC_SCRATCH_RAM_ADDR_DEFAULT                                          0x00000000
#define regVPEC_QUEUE_RESET_REQ_DEFAULT                                           0x00000000
#define regVPEC_PERFCNT_PERFCOUNTER0_CFG_DEFAULT                                  0x0000ffff
#define regVPEC_PERFCNT_PERFCOUNTER1_CFG_DEFAULT                                  0x0000ffff
#define regVPEC_PERFCNT_PERFCOUNTER_RSLT_CNTL_DEFAULT                             0x04000000
#define regVPEC_PERFCNT_MISC_CNTL_DEFAULT                                         0x00010000
#define regVPEC_PERFCNT_PERFCOUNTER_LO_DEFAULT                                    0x00000000
#define regVPEC_PERFCNT_PERFCOUNTER_HI_DEFAULT                                    0x00000000
#define regVPEC_DEBUG_INDEX_DEFAULT                                               0x00000000
#define regVPEC_DEBUG_DATA_DEFAULT                                                0x00000000
#define regVPEC_CRC_CTRL_DEFAULT                                                  0x00000000
#define regVPEC_CRC_DATA_DEFAULT                                                  0x00000000
#define regVPEC_PUB_DUMMY0_DEFAULT                                                0x00000000
#define regVPEC_PUB_DUMMY1_DEFAULT                                                0x00000000
#define regVPEC_PUB_DUMMY2_DEFAULT                                                0x00000000
#define regVPEC_PUB_DUMMY3_DEFAULT                                                0x00000000
#define regVPEC_PUB_DUMMY4_DEFAULT                                                0x00000000
#define regVPEC_PUB_DUMMY5_DEFAULT                                                0x00000000
#define regVPEC_PUB_DUMMY6_DEFAULT                                                0x00000000
#define regVPEC_PUB_DUMMY7_DEFAULT                                                0x00000000
#define regVPEC_UCODE1_CHECKSUM_DEFAULT                                           0x00000000
#define regVPEC_VERSION_DEFAULT                                                   0x00000601
#define regVPEC_UCODE_CHECKSUM_DEFAULT                                            0x00000000
#define regVPEC_CLOCK_GATING_STATUS_DEFAULT                                       0x00000000
#define regVPEC_RB_RPTR_FETCH_DEFAULT                                             0x00000000
#define regVPEC_RB_RPTR_FETCH_HI_DEFAULT                                          0x00000000
#define regVPEC_IB_OFFSET_FETCH_DEFAULT                                           0x00000000
#define regVPEC_CMDIB_OFFSET_FETCH_DEFAULT                                        0x00000000
#define regVPEC_ATOMIC_PREOP_LO_DEFAULT                                           0x00000000
#define regVPEC_ATOMIC_PREOP_HI_DEFAULT                                           0x00000000
#define regVPEC_CE_BUSY_DEFAULT                                                   0x00000000
#define regVPEC_F32_COUNTER_DEFAULT                                               0x00000000
#define regVPEC_HOLE_ADDR_LO_DEFAULT                                              0x00000000
#define regVPEC_HOLE_ADDR_HI_DEFAULT                                              0x00000000
#define regVPEC_ERROR_LOG_DEFAULT                                                 0x0000000f
#define regVPEC_INT_STATUS_DEFAULT                                                0x00000000
#define regVPEC_STATUS_DEFAULT                                                    0x46dee557
#define regVPEC_STATUS1_DEFAULT                                                   0x0003ffff
#define regVPEC_STATUS2_DEFAULT                                                   0x00000000
#define regVPEC_STATUS3_DEFAULT                                                   0x03f00000
#define regVPEC_STATUS4_DEFAULT                                                   0x00000001
#define regVPEC_STATUS5_DEFAULT                                                   0x00000000
#define regVPEC_STATUS6_DEFAULT                                                   0x00000000
#define regVPEC_STATUS7_DEFAULT                                                   0x00000000
#define regVPEC_INST_DEFAULT                                                      0x00000000
#define regVPEC_QUEUE_STATUS0_DEFAULT                                             0x22222222
#define regVPEC_QUEUE_HANG_STATUS_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_RB_CNTL_DEFAULT                                            0x00040800
#define regVPEC_QUEUE0_SCHEDULE_CNTL_DEFAULT                                      0x00000000
#define regVPEC_QUEUE0_RB_BASE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE0_RB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_RB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE0_RB_RPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_RB_WPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE0_RB_WPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_RB_RPTR_ADDR_HI_DEFAULT                                    0x00000000
#define regVPEC_QUEUE0_RB_RPTR_ADDR_LO_DEFAULT                                    0x00000000
#define regVPEC_QUEUE0_RB_AQL_CNTL_DEFAULT                                        0x00004000
#define regVPEC_QUEUE0_MINOR_PTR_UPDATE_DEFAULT                                   0x00000000
#define regVPEC_QUEUE0_CD_INFO_DEFAULT                                            0x00000000
#define regVPEC_QUEUE0_RB_PREEMPT_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_SKIP_CNTL_DEFAULT                                          0x00000000
#define regVPEC_QUEUE0_DOORBELL_DEFAULT                                           0x00000000
#define regVPEC_QUEUE0_DOORBELL_OFFSET_DEFAULT                                    0x00000000
#define regVPEC_QUEUE0_DUMMY0_DEFAULT                                             0x0000000f
#define regVPEC_QUEUE0_DUMMY1_DEFAULT                                             0x00000000
#define regVPEC_QUEUE0_DUMMY2_DEFAULT                                             0x00000000
#define regVPEC_QUEUE0_DUMMY3_DEFAULT                                             0x00000000
#define regVPEC_QUEUE0_DUMMY4_DEFAULT                                             0x00000000
#define regVPEC_QUEUE0_IB_CNTL_DEFAULT                                            0x00000100
#define regVPEC_QUEUE0_IB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE0_IB_OFFSET_DEFAULT                                          0x00000000
#define regVPEC_QUEUE0_IB_BASE_LO_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_IB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_IB_SIZE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE0_CMDIB_CNTL_DEFAULT                                         0x00000101
#define regVPEC_QUEUE0_CMDIB_RPTR_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_CMDIB_OFFSET_DEFAULT                                       0x00000000
#define regVPEC_QUEUE0_CMDIB_BASE_LO_DEFAULT                                      0x00000000
#define regVPEC_QUEUE0_CMDIB_BASE_HI_DEFAULT                                      0x00000000
#define regVPEC_QUEUE0_CMDIB_SIZE_DEFAULT                                         0x00000000
#define regVPEC_QUEUE0_CSA_ADDR_LO_DEFAULT                                        0x00000000
#define regVPEC_QUEUE0_CSA_ADDR_HI_DEFAULT                                        0x00000000
#define regVPEC_QUEUE0_CONTEXT_STATUS_DEFAULT                                     0x00000804
#define regVPEC_QUEUE0_DOORBELL_LOG_DEFAULT                                       0x00000000
#define regVPEC_QUEUE0_IB_SUB_REMAIN_DEFAULT                                      0x00000000
#define regVPEC_QUEUE0_PREEMPT_DEFAULT                                            0x00000000
#define regVPEC_QUEUE1_RB_CNTL_DEFAULT                                            0x00040800
#define regVPEC_QUEUE1_SCHEDULE_CNTL_DEFAULT                                      0x00000000
#define regVPEC_QUEUE1_RB_BASE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE1_RB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE1_RB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE1_RB_RPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE1_RB_WPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE1_RB_WPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE1_RB_RPTR_ADDR_HI_DEFAULT                                    0x00000000
#define regVPEC_QUEUE1_RB_RPTR_ADDR_LO_DEFAULT                                    0x00000000
#define regVPEC_QUEUE1_RB_AQL_CNTL_DEFAULT                                        0x00004000
#define regVPEC_QUEUE1_MINOR_PTR_UPDATE_DEFAULT                                   0x00000000
#define regVPEC_QUEUE1_CD_INFO_DEFAULT                                            0x00000000
#define regVPEC_QUEUE1_RB_PREEMPT_DEFAULT                                         0x00000000
#define regVPEC_QUEUE1_SKIP_CNTL_DEFAULT                                          0x00000000
#define regVPEC_QUEUE1_DOORBELL_DEFAULT                                           0x00000000
#define regVPEC_QUEUE1_DOORBELL_OFFSET_DEFAULT                                    0x00000000
#define regVPEC_QUEUE1_DUMMY0_DEFAULT                                             0x0000000f
#define regVPEC_QUEUE1_DUMMY1_DEFAULT                                             0x00000000
#define regVPEC_QUEUE1_DUMMY2_DEFAULT                                             0x00000000
#define regVPEC_QUEUE1_DUMMY3_DEFAULT                                             0x00000000
#define regVPEC_QUEUE1_DUMMY4_DEFAULT                                             0x00000000
#define regVPEC_QUEUE1_IB_CNTL_DEFAULT                                            0x00000100
#define regVPEC_QUEUE1_IB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE1_IB_OFFSET_DEFAULT                                          0x00000000
#define regVPEC_QUEUE1_IB_BASE_LO_DEFAULT                                         0x00000000
#define regVPEC_QUEUE1_IB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE1_IB_SIZE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE1_CMDIB_CNTL_DEFAULT                                         0x00000101
#define regVPEC_QUEUE1_CMDIB_RPTR_DEFAULT                                         0x00000000
#define regVPEC_QUEUE1_CMDIB_OFFSET_DEFAULT                                       0x00000000
#define regVPEC_QUEUE1_CMDIB_BASE_LO_DEFAULT                                      0x00000000
#define regVPEC_QUEUE1_CMDIB_BASE_HI_DEFAULT                                      0x00000000
#define regVPEC_QUEUE1_CMDIB_SIZE_DEFAULT                                         0x00000000
#define regVPEC_QUEUE1_CSA_ADDR_LO_DEFAULT                                        0x00000000
#define regVPEC_QUEUE1_CSA_ADDR_HI_DEFAULT                                        0x00000000
#define regVPEC_QUEUE1_CONTEXT_STATUS_DEFAULT                                     0x00000804
#define regVPEC_QUEUE1_DOORBELL_LOG_DEFAULT                                       0x00000000
#define regVPEC_QUEUE1_IB_SUB_REMAIN_DEFAULT                                      0x00000000
#define regVPEC_QUEUE1_PREEMPT_DEFAULT                                            0x00000000
#define regVPEC_QUEUE2_RB_CNTL_DEFAULT                                            0x00040800
#define regVPEC_QUEUE2_SCHEDULE_CNTL_DEFAULT                                      0x00000000
#define regVPEC_QUEUE2_RB_BASE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE2_RB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE2_RB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE2_RB_RPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE2_RB_WPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE2_RB_WPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE2_RB_RPTR_ADDR_HI_DEFAULT                                    0x00000000
#define regVPEC_QUEUE2_RB_RPTR_ADDR_LO_DEFAULT                                    0x00000000
#define regVPEC_QUEUE2_RB_AQL_CNTL_DEFAULT                                        0x00004000
#define regVPEC_QUEUE2_MINOR_PTR_UPDATE_DEFAULT                                   0x00000000
#define regVPEC_QUEUE2_CD_INFO_DEFAULT                                            0x00000000
#define regVPEC_QUEUE2_RB_PREEMPT_DEFAULT                                         0x00000000
#define regVPEC_QUEUE2_SKIP_CNTL_DEFAULT                                          0x00000000
#define regVPEC_QUEUE2_DOORBELL_DEFAULT                                           0x00000000
#define regVPEC_QUEUE2_DOORBELL_OFFSET_DEFAULT                                    0x00000000
#define regVPEC_QUEUE2_DUMMY0_DEFAULT                                             0x0000000f
#define regVPEC_QUEUE2_DUMMY1_DEFAULT                                             0x00000000
#define regVPEC_QUEUE2_DUMMY2_DEFAULT                                             0x00000000
#define regVPEC_QUEUE2_DUMMY3_DEFAULT                                             0x00000000
#define regVPEC_QUEUE2_DUMMY4_DEFAULT                                             0x00000000
#define regVPEC_QUEUE2_IB_CNTL_DEFAULT                                            0x00000100
#define regVPEC_QUEUE2_IB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE2_IB_OFFSET_DEFAULT                                          0x00000000
#define regVPEC_QUEUE2_IB_BASE_LO_DEFAULT                                         0x00000000
#define regVPEC_QUEUE2_IB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE2_IB_SIZE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE2_CMDIB_CNTL_DEFAULT                                         0x00000101
#define regVPEC_QUEUE2_CMDIB_RPTR_DEFAULT                                         0x00000000
#define regVPEC_QUEUE2_CMDIB_OFFSET_DEFAULT                                       0x00000000
#define regVPEC_QUEUE2_CMDIB_BASE_LO_DEFAULT                                      0x00000000
#define regVPEC_QUEUE2_CMDIB_BASE_HI_DEFAULT                                      0x00000000
#define regVPEC_QUEUE2_CMDIB_SIZE_DEFAULT                                         0x00000000
#define regVPEC_QUEUE2_CSA_ADDR_LO_DEFAULT                                        0x00000000
#define regVPEC_QUEUE2_CSA_ADDR_HI_DEFAULT                                        0x00000000
#define regVPEC_QUEUE2_CONTEXT_STATUS_DEFAULT                                     0x00000804
#define regVPEC_QUEUE2_DOORBELL_LOG_DEFAULT                                       0x00000000
#define regVPEC_QUEUE2_IB_SUB_REMAIN_DEFAULT                                      0x00000000
#define regVPEC_QUEUE2_PREEMPT_DEFAULT                                            0x00000000
#define regVPEC_QUEUE3_RB_CNTL_DEFAULT                                            0x00040800
#define regVPEC_QUEUE3_SCHEDULE_CNTL_DEFAULT                                      0x00000000
#define regVPEC_QUEUE3_RB_BASE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE3_RB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE3_RB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE3_RB_RPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE3_RB_WPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE3_RB_WPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE3_RB_RPTR_ADDR_HI_DEFAULT                                    0x00000000
#define regVPEC_QUEUE3_RB_RPTR_ADDR_LO_DEFAULT                                    0x00000000
#define regVPEC_QUEUE3_RB_AQL_CNTL_DEFAULT                                        0x00004000
#define regVPEC_QUEUE3_MINOR_PTR_UPDATE_DEFAULT                                   0x00000000
#define regVPEC_QUEUE3_CD_INFO_DEFAULT                                            0x00000000
#define regVPEC_QUEUE3_RB_PREEMPT_DEFAULT                                         0x00000000
#define regVPEC_QUEUE3_SKIP_CNTL_DEFAULT                                          0x00000000
#define regVPEC_QUEUE3_DOORBELL_DEFAULT                                           0x00000000
#define regVPEC_QUEUE3_DOORBELL_OFFSET_DEFAULT                                    0x00000000
#define regVPEC_QUEUE3_DUMMY0_DEFAULT                                             0x0000000f
#define regVPEC_QUEUE3_DUMMY1_DEFAULT                                             0x00000000
#define regVPEC_QUEUE3_DUMMY2_DEFAULT                                             0x00000000
#define regVPEC_QUEUE3_DUMMY3_DEFAULT                                             0x00000000
#define regVPEC_QUEUE3_DUMMY4_DEFAULT                                             0x00000000
#define regVPEC_QUEUE3_IB_CNTL_DEFAULT                                            0x00000100
#define regVPEC_QUEUE3_IB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE3_IB_OFFSET_DEFAULT                                          0x00000000
#define regVPEC_QUEUE3_IB_BASE_LO_DEFAULT                                         0x00000000
#define regVPEC_QUEUE3_IB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE3_IB_SIZE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE3_CMDIB_CNTL_DEFAULT                                         0x00000101
#define regVPEC_QUEUE3_CMDIB_RPTR_DEFAULT                                         0x00000000
#define regVPEC_QUEUE3_CMDIB_OFFSET_DEFAULT                                       0x00000000
#define regVPEC_QUEUE3_CMDIB_BASE_LO_DEFAULT                                      0x00000000
#define regVPEC_QUEUE3_CMDIB_BASE_HI_DEFAULT                                      0x00000000
#define regVPEC_QUEUE3_CMDIB_SIZE_DEFAULT                                         0x00000000
#define regVPEC_QUEUE3_CSA_ADDR_LO_DEFAULT                                        0x00000000
#define regVPEC_QUEUE3_CSA_ADDR_HI_DEFAULT                                        0x00000000
#define regVPEC_QUEUE3_CONTEXT_STATUS_DEFAULT                                     0x00000804
#define regVPEC_QUEUE3_DOORBELL_LOG_DEFAULT                                       0x00000000
#define regVPEC_QUEUE3_IB_SUB_REMAIN_DEFAULT                                      0x00000000
#define regVPEC_QUEUE3_PREEMPT_DEFAULT                                            0x00000000
#define regVPEC_QUEUE4_RB_CNTL_DEFAULT                                            0x00040800
#define regVPEC_QUEUE4_SCHEDULE_CNTL_DEFAULT                                      0x00000000
#define regVPEC_QUEUE4_RB_BASE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE4_RB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE4_RB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE4_RB_RPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE4_RB_WPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE4_RB_WPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE4_RB_RPTR_ADDR_HI_DEFAULT                                    0x00000000
#define regVPEC_QUEUE4_RB_RPTR_ADDR_LO_DEFAULT                                    0x00000000
#define regVPEC_QUEUE4_RB_AQL_CNTL_DEFAULT                                        0x00004000
#define regVPEC_QUEUE4_MINOR_PTR_UPDATE_DEFAULT                                   0x00000000
#define regVPEC_QUEUE4_CD_INFO_DEFAULT                                            0x00000000
#define regVPEC_QUEUE4_RB_PREEMPT_DEFAULT                                         0x00000000
#define regVPEC_QUEUE4_SKIP_CNTL_DEFAULT                                          0x00000000
#define regVPEC_QUEUE4_DOORBELL_DEFAULT                                           0x00000000
#define regVPEC_QUEUE4_DOORBELL_OFFSET_DEFAULT                                    0x00000000
#define regVPEC_QUEUE4_DUMMY0_DEFAULT                                             0x0000000f
#define regVPEC_QUEUE4_DUMMY1_DEFAULT                                             0x00000000
#define regVPEC_QUEUE4_DUMMY2_DEFAULT                                             0x00000000
#define regVPEC_QUEUE4_DUMMY3_DEFAULT                                             0x00000000
#define regVPEC_QUEUE4_DUMMY4_DEFAULT                                             0x00000000
#define regVPEC_QUEUE4_IB_CNTL_DEFAULT                                            0x00000100
#define regVPEC_QUEUE4_IB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE4_IB_OFFSET_DEFAULT                                          0x00000000
#define regVPEC_QUEUE4_IB_BASE_LO_DEFAULT                                         0x00000000
#define regVPEC_QUEUE4_IB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE4_IB_SIZE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE4_CMDIB_CNTL_DEFAULT                                         0x00000101
#define regVPEC_QUEUE4_CMDIB_RPTR_DEFAULT                                         0x00000000
#define regVPEC_QUEUE4_CMDIB_OFFSET_DEFAULT                                       0x00000000
#define regVPEC_QUEUE4_CMDIB_BASE_LO_DEFAULT                                      0x00000000
#define regVPEC_QUEUE4_CMDIB_BASE_HI_DEFAULT                                      0x00000000
#define regVPEC_QUEUE4_CMDIB_SIZE_DEFAULT                                         0x00000000
#define regVPEC_QUEUE4_CSA_ADDR_LO_DEFAULT                                        0x00000000
#define regVPEC_QUEUE4_CSA_ADDR_HI_DEFAULT                                        0x00000000
#define regVPEC_QUEUE4_CONTEXT_STATUS_DEFAULT                                     0x00000804
#define regVPEC_QUEUE4_DOORBELL_LOG_DEFAULT                                       0x00000000
#define regVPEC_QUEUE4_IB_SUB_REMAIN_DEFAULT                                      0x00000000
#define regVPEC_QUEUE4_PREEMPT_DEFAULT                                            0x00000000
#define regVPEC_QUEUE5_RB_CNTL_DEFAULT                                            0x00040800
#define regVPEC_QUEUE5_SCHEDULE_CNTL_DEFAULT                                      0x00000000
#define regVPEC_QUEUE5_RB_BASE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE5_RB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE5_RB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE5_RB_RPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE5_RB_WPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE5_RB_WPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE5_RB_RPTR_ADDR_HI_DEFAULT                                    0x00000000
#define regVPEC_QUEUE5_RB_RPTR_ADDR_LO_DEFAULT                                    0x00000000
#define regVPEC_QUEUE5_RB_AQL_CNTL_DEFAULT                                        0x00004000
#define regVPEC_QUEUE5_MINOR_PTR_UPDATE_DEFAULT                                   0x00000000
#define regVPEC_QUEUE5_CD_INFO_DEFAULT                                            0x00000000
#define regVPEC_QUEUE5_RB_PREEMPT_DEFAULT                                         0x00000000
#define regVPEC_QUEUE5_SKIP_CNTL_DEFAULT                                          0x00000000
#define regVPEC_QUEUE5_DOORBELL_DEFAULT                                           0x00000000
#define regVPEC_QUEUE5_DOORBELL_OFFSET_DEFAULT                                    0x00000000
#define regVPEC_QUEUE5_DUMMY0_DEFAULT                                             0x0000000f
#define regVPEC_QUEUE5_DUMMY1_DEFAULT                                             0x00000000
#define regVPEC_QUEUE5_DUMMY2_DEFAULT                                             0x00000000
#define regVPEC_QUEUE5_DUMMY3_DEFAULT                                             0x00000000
#define regVPEC_QUEUE5_DUMMY4_DEFAULT                                             0x00000000
#define regVPEC_QUEUE5_IB_CNTL_DEFAULT                                            0x00000100
#define regVPEC_QUEUE5_IB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE5_IB_OFFSET_DEFAULT                                          0x00000000
#define regVPEC_QUEUE5_IB_BASE_LO_DEFAULT                                         0x00000000
#define regVPEC_QUEUE5_IB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE5_IB_SIZE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE5_CMDIB_CNTL_DEFAULT                                         0x00000101
#define regVPEC_QUEUE5_CMDIB_RPTR_DEFAULT                                         0x00000000
#define regVPEC_QUEUE5_CMDIB_OFFSET_DEFAULT                                       0x00000000
#define regVPEC_QUEUE5_CMDIB_BASE_LO_DEFAULT                                      0x00000000
#define regVPEC_QUEUE5_CMDIB_BASE_HI_DEFAULT                                      0x00000000
#define regVPEC_QUEUE5_CMDIB_SIZE_DEFAULT                                         0x00000000
#define regVPEC_QUEUE5_CSA_ADDR_LO_DEFAULT                                        0x00000000
#define regVPEC_QUEUE5_CSA_ADDR_HI_DEFAULT                                        0x00000000
#define regVPEC_QUEUE5_CONTEXT_STATUS_DEFAULT                                     0x00000804
#define regVPEC_QUEUE5_DOORBELL_LOG_DEFAULT                                       0x00000000
#define regVPEC_QUEUE5_IB_SUB_REMAIN_DEFAULT                                      0x00000000
#define regVPEC_QUEUE5_PREEMPT_DEFAULT                                            0x00000000
#define regVPEC_QUEUE6_RB_CNTL_DEFAULT                                            0x00040800
#define regVPEC_QUEUE6_SCHEDULE_CNTL_DEFAULT                                      0x00000000
#define regVPEC_QUEUE6_RB_BASE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE6_RB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE6_RB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE6_RB_RPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE6_RB_WPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE6_RB_WPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE6_RB_RPTR_ADDR_HI_DEFAULT                                    0x00000000
#define regVPEC_QUEUE6_RB_RPTR_ADDR_LO_DEFAULT                                    0x00000000
#define regVPEC_QUEUE6_RB_AQL_CNTL_DEFAULT                                        0x00004000
#define regVPEC_QUEUE6_MINOR_PTR_UPDATE_DEFAULT                                   0x00000000
#define regVPEC_QUEUE6_CD_INFO_DEFAULT                                            0x00000000
#define regVPEC_QUEUE6_RB_PREEMPT_DEFAULT                                         0x00000000
#define regVPEC_QUEUE6_SKIP_CNTL_DEFAULT                                          0x00000000
#define regVPEC_QUEUE6_DOORBELL_DEFAULT                                           0x00000000
#define regVPEC_QUEUE6_DOORBELL_OFFSET_DEFAULT                                    0x00000000
#define regVPEC_QUEUE6_DUMMY0_DEFAULT                                             0x0000000f
#define regVPEC_QUEUE6_DUMMY1_DEFAULT                                             0x00000000
#define regVPEC_QUEUE6_DUMMY2_DEFAULT                                             0x00000000
#define regVPEC_QUEUE6_DUMMY3_DEFAULT                                             0x00000000
#define regVPEC_QUEUE6_DUMMY4_DEFAULT                                             0x00000000
#define regVPEC_QUEUE6_IB_CNTL_DEFAULT                                            0x00000100
#define regVPEC_QUEUE6_IB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE6_IB_OFFSET_DEFAULT                                          0x00000000
#define regVPEC_QUEUE6_IB_BASE_LO_DEFAULT                                         0x00000000
#define regVPEC_QUEUE6_IB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE6_IB_SIZE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE6_CMDIB_CNTL_DEFAULT                                         0x00000101
#define regVPEC_QUEUE6_CMDIB_RPTR_DEFAULT                                         0x00000000
#define regVPEC_QUEUE6_CMDIB_OFFSET_DEFAULT                                       0x00000000
#define regVPEC_QUEUE6_CMDIB_BASE_LO_DEFAULT                                      0x00000000
#define regVPEC_QUEUE6_CMDIB_BASE_HI_DEFAULT                                      0x00000000
#define regVPEC_QUEUE6_CMDIB_SIZE_DEFAULT                                         0x00000000
#define regVPEC_QUEUE6_CSA_ADDR_LO_DEFAULT                                        0x00000000
#define regVPEC_QUEUE6_CSA_ADDR_HI_DEFAULT                                        0x00000000
#define regVPEC_QUEUE6_CONTEXT_STATUS_DEFAULT                                     0x00000804
#define regVPEC_QUEUE6_DOORBELL_LOG_DEFAULT                                       0x00000000
#define regVPEC_QUEUE6_IB_SUB_REMAIN_DEFAULT                                      0x00000000
#define regVPEC_QUEUE6_PREEMPT_DEFAULT                                            0x00000000
#define regVPEC_QUEUE7_RB_CNTL_DEFAULT                                            0x00040800
#define regVPEC_QUEUE7_SCHEDULE_CNTL_DEFAULT                                      0x00000000
#define regVPEC_QUEUE7_RB_BASE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE7_RB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE7_RB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE7_RB_RPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE7_RB_WPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE7_RB_WPTR_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE7_RB_RPTR_ADDR_HI_DEFAULT                                    0x00000000
#define regVPEC_QUEUE7_RB_RPTR_ADDR_LO_DEFAULT                                    0x00000000
#define regVPEC_QUEUE7_RB_AQL_CNTL_DEFAULT                                        0x00004000
#define regVPEC_QUEUE7_MINOR_PTR_UPDATE_DEFAULT                                   0x00000000
#define regVPEC_QUEUE7_CD_INFO_DEFAULT                                            0x00000000
#define regVPEC_QUEUE7_RB_PREEMPT_DEFAULT                                         0x00000000
#define regVPEC_QUEUE7_SKIP_CNTL_DEFAULT                                          0x00000000
#define regVPEC_QUEUE7_DOORBELL_DEFAULT                                           0x00000000
#define regVPEC_QUEUE7_DOORBELL_OFFSET_DEFAULT                                    0x00000000
#define regVPEC_QUEUE7_DUMMY0_DEFAULT                                             0x0000000f
#define regVPEC_QUEUE7_DUMMY1_DEFAULT                                             0x00000000
#define regVPEC_QUEUE7_DUMMY2_DEFAULT                                             0x00000000
#define regVPEC_QUEUE7_DUMMY3_DEFAULT                                             0x00000000
#define regVPEC_QUEUE7_DUMMY4_DEFAULT                                             0x00000000
#define regVPEC_QUEUE7_IB_CNTL_DEFAULT                                            0x00000100
#define regVPEC_QUEUE7_IB_RPTR_DEFAULT                                            0x00000000
#define regVPEC_QUEUE7_IB_OFFSET_DEFAULT                                          0x00000000
#define regVPEC_QUEUE7_IB_BASE_LO_DEFAULT                                         0x00000000
#define regVPEC_QUEUE7_IB_BASE_HI_DEFAULT                                         0x00000000
#define regVPEC_QUEUE7_IB_SIZE_DEFAULT                                            0x00000000
#define regVPEC_QUEUE7_CMDIB_CNTL_DEFAULT                                         0x00000101
#define regVPEC_QUEUE7_CMDIB_RPTR_DEFAULT                                         0x00000000
#define regVPEC_QUEUE7_CMDIB_OFFSET_DEFAULT                                       0x00000000
#define regVPEC_QUEUE7_CMDIB_BASE_LO_DEFAULT                                      0x00000000
#define regVPEC_QUEUE7_CMDIB_BASE_HI_DEFAULT                                      0x00000000
#define regVPEC_QUEUE7_CMDIB_SIZE_DEFAULT                                         0x00000000
#define regVPEC_QUEUE7_CSA_ADDR_LO_DEFAULT                                        0x00000000
#define regVPEC_QUEUE7_CSA_ADDR_HI_DEFAULT                                        0x00000000
#define regVPEC_QUEUE7_CONTEXT_STATUS_DEFAULT                                     0x00000804
#define regVPEC_QUEUE7_DOORBELL_LOG_DEFAULT                                       0x00000000
#define regVPEC_QUEUE7_IB_SUB_REMAIN_DEFAULT                                      0x00000000
#define regVPEC_QUEUE7_PREEMPT_DEFAULT                                            0x00000000


// addressBlock: vpe_vpep_vpdpp0_dispdec_vpcnvc_cfg_dispdec
#define regVPCNVC_SURFACE_PIXEL_FORMAT_DEFAULT                                    0x00000008
#define regVPCNVC_FORMAT_CONTROL_DEFAULT                                          0x00000000
#define regVPCNVC_FCNV_FP_BIAS_R_DEFAULT                                          0x00000000
#define regVPCNVC_FCNV_FP_BIAS_G_DEFAULT                                          0x00000000
#define regVPCNVC_FCNV_FP_BIAS_B_DEFAULT                                          0x00000000
#define regVPCNVC_FCNV_FP_SCALE_R_DEFAULT                                         0x0001f000
#define regVPCNVC_FCNV_FP_SCALE_G_DEFAULT                                         0x0001f000
#define regVPCNVC_FCNV_FP_SCALE_B_DEFAULT                                         0x0001f000
#define regVPCNVC_COLOR_KEYER_CONTROL_DEFAULT                                     0x00000000
#define regVPCNVC_COLOR_KEYER_ALPHA_DEFAULT                                       0x00000000
#define regVPCNVC_COLOR_KEYER_RED_DEFAULT                                         0x00000000
#define regVPCNVC_COLOR_KEYER_GREEN_DEFAULT                                       0x00000000
#define regVPCNVC_COLOR_KEYER_BLUE_DEFAULT                                        0x00000000
#define regVPCNVC_ALPHA_2BIT_LUT_DEFAULT                                          0xffaa5500
#define regVPCNVC_PRE_DEALPHA_DEFAULT                                             0x00000000
#define regVPCNVC_PRE_CSC_MODE_DEFAULT                                            0x00000000
#define regVPCNVC_PRE_CSC_C11_C12_DEFAULT                                         0x00002000
#define regVPCNVC_PRE_CSC_C13_C14_DEFAULT                                         0x00000000
#define regVPCNVC_PRE_CSC_C21_C22_DEFAULT                                         0x20000000
#define regVPCNVC_PRE_CSC_C23_C24_DEFAULT                                         0x00000000
#define regVPCNVC_PRE_CSC_C31_C32_DEFAULT                                         0x00000000
#define regVPCNVC_PRE_CSC_C33_C34_DEFAULT                                         0x00002000
#define regVPCNVC_COEF_FORMAT_DEFAULT                                             0x00000000
#define regVPCNVC_PRE_DEGAM_DEFAULT                                               0x00000000
#define regVPCNVC_PRE_REALPHA_DEFAULT                                             0x00000000
#define regVPCNVC_CFG_TEST_DEBUG_INDEX_DEFAULT                                    0x00000000
#define regVPCNVC_CFG_TEST_DEBUG_DATA_DEFAULT                                     0x00000000


// addressBlock: vpe_vpep_vpdpp0_dispdec_vpdscl_dispdec
#define regVPDSCL_COEF_RAM_TAP_SELECT_DEFAULT                                     0x00000000
#define regVPDSCL_COEF_RAM_TAP_DATA_DEFAULT                                       0x00000000
#define regVPDSCL_MODE_DEFAULT                                                    0x00000000
#define regVPDSCL_TAP_CONTROL_DEFAULT                                             0x00000000
#define regVPDSCL_CONTROL_DEFAULT                                                 0x00000000
#define regVPDSCL_2TAP_CONTROL_DEFAULT                                            0x01000100
#define regVPDSCL_MANUAL_REPLICATE_CONTROL_DEFAULT                                0x00000000
#define regVPDSCL_HORZ_FILTER_SCALE_RATIO_DEFAULT                                 0x00000000
#define regVPDSCL_HORZ_FILTER_INIT_DEFAULT                                        0x01000000
#define regVPDSCL_HORZ_FILTER_SCALE_RATIO_C_DEFAULT                               0x00000000
#define regVPDSCL_HORZ_FILTER_INIT_C_DEFAULT                                      0x01000000
#define regVPDSCL_VERT_FILTER_SCALE_RATIO_DEFAULT                                 0x00000000
#define regVPDSCL_VERT_FILTER_INIT_DEFAULT                                        0x01000000
#define regVPDSCL_VERT_FILTER_INIT_BOT_DEFAULT                                    0x01000000
#define regVPDSCL_VERT_FILTER_SCALE_RATIO_C_DEFAULT                               0x00000000
#define regVPDSCL_VERT_FILTER_INIT_C_DEFAULT                                      0x01000000
#define regVPDSCL_VERT_FILTER_INIT_BOT_C_DEFAULT                                  0x01000000
#define regVPDSCL_BLACK_COLOR_DEFAULT                                             0x3c000000
#define regVPDSCL_UPDATE_DEFAULT                                                  0x00000000
#define regVPDSCL_AUTOCAL_DEFAULT                                                 0x00000000
#define regVPDSCL_EXT_OVERSCAN_LEFT_RIGHT_DEFAULT                                 0x00000000
#define regVPDSCL_EXT_OVERSCAN_TOP_BOTTOM_DEFAULT                                 0x00000000
#define regVPOTG_H_BLANK_DEFAULT                                                  0x00000000
#define regVPOTG_V_BLANK_DEFAULT                                                  0x00000000
#define regVPDSCL_RECOUT_START_DEFAULT                                            0x00000000
#define regVPDSCL_RECOUT_SIZE_DEFAULT                                             0x00000000
#define regVPMPC_SIZE_DEFAULT                                                     0x00000000
#define regVPLB_DATA_FORMAT_DEFAULT                                               0x00000000
#define regVPLB_MEMORY_CTRL_DEFAULT                                               0x00003f00
#define regVPLB_V_COUNTER_DEFAULT                                                 0x00000000
#define regVPDSCL_MEM_PWR_CTRL_DEFAULT                                            0x00000000
#define regVPDSCL_MEM_PWR_STATUS_DEFAULT                                          0x00000000
#define regVPDSCL_DEBUG_DEFAULT                                                   0x00000000
#define regVPDSCL_TEST_DEBUG_INDEX_DEFAULT                                        0x00000000
#define regVPDSCL_TEST_DEBUG_DATA_DEFAULT                                         0x00000000


// addressBlock: vpe_vpep_vpdpp0_dispdec_vpcm_dispdec
#define regVPCM_CONTROL_DEFAULT                                                   0x00000000
#define regVPCM_POST_CSC_CONTROL_DEFAULT                                          0x00000000
#define regVPCM_POST_CSC_C11_C12_DEFAULT                                          0x00002000
#define regVPCM_POST_CSC_C13_C14_DEFAULT                                          0x00000000
#define regVPCM_POST_CSC_C21_C22_DEFAULT                                          0x20000000
#define regVPCM_POST_CSC_C23_C24_DEFAULT                                          0x00000000
#define regVPCM_POST_CSC_C31_C32_DEFAULT                                          0x00000000
#define regVPCM_POST_CSC_C33_C34_DEFAULT                                          0x00002000
#define regVPCM_GAMUT_REMAP_CONTROL_DEFAULT                                       0x00000000
#define regVPCM_GAMUT_REMAP_C11_C12_DEFAULT                                       0x00002000
#define regVPCM_GAMUT_REMAP_C13_C14_DEFAULT                                       0x00000000
#define regVPCM_GAMUT_REMAP_C21_C22_DEFAULT                                       0x20000000
#define regVPCM_GAMUT_REMAP_C23_C24_DEFAULT                                       0x00000000
#define regVPCM_GAMUT_REMAP_C31_C32_DEFAULT                                       0x00000000
#define regVPCM_GAMUT_REMAP_C33_C34_DEFAULT                                       0x00002000
#define regVPCM_BIAS_CR_R_DEFAULT                                                 0x00000000
#define regVPCM_BIAS_Y_G_CB_B_DEFAULT                                             0x00000000
#define regVPCM_GAMCOR_CONTROL_DEFAULT                                            0x00000000
#define regVPCM_GAMCOR_LUT_INDEX_DEFAULT                                          0x00000000
#define regVPCM_GAMCOR_LUT_DATA_DEFAULT                                           0x00000000
#define regVPCM_GAMCOR_LUT_CONTROL_DEFAULT                                        0x00000007
#define regVPCM_GAMCOR_RAMA_START_CNTL_B_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_START_CNTL_G_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_START_CNTL_R_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_START_SLOPE_CNTL_B_DEFAULT                            0x00000000
#define regVPCM_GAMCOR_RAMA_START_SLOPE_CNTL_G_DEFAULT                            0x00000000
#define regVPCM_GAMCOR_RAMA_START_SLOPE_CNTL_R_DEFAULT                            0x00000000
#define regVPCM_GAMCOR_RAMA_START_BASE_CNTL_B_DEFAULT                             0x00000000
#define regVPCM_GAMCOR_RAMA_START_BASE_CNTL_G_DEFAULT                             0x00000000
#define regVPCM_GAMCOR_RAMA_START_BASE_CNTL_R_DEFAULT                             0x00000000
#define regVPCM_GAMCOR_RAMA_END_CNTL1_B_DEFAULT                                   0x00000000
#define regVPCM_GAMCOR_RAMA_END_CNTL2_B_DEFAULT                                   0x00000000
#define regVPCM_GAMCOR_RAMA_END_CNTL1_G_DEFAULT                                   0x00000000
#define regVPCM_GAMCOR_RAMA_END_CNTL2_G_DEFAULT                                   0x00000000
#define regVPCM_GAMCOR_RAMA_END_CNTL1_R_DEFAULT                                   0x00000000
#define regVPCM_GAMCOR_RAMA_END_CNTL2_R_DEFAULT                                   0x00000000
#define regVPCM_GAMCOR_RAMA_OFFSET_B_DEFAULT                                      0x00000000
#define regVPCM_GAMCOR_RAMA_OFFSET_G_DEFAULT                                      0x00000000
#define regVPCM_GAMCOR_RAMA_OFFSET_R_DEFAULT                                      0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_0_1_DEFAULT                                    0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_2_3_DEFAULT                                    0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_4_5_DEFAULT                                    0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_6_7_DEFAULT                                    0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_8_9_DEFAULT                                    0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_10_11_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_12_13_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_14_15_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_16_17_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_18_19_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_20_21_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_22_23_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_24_25_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_26_27_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_28_29_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_30_31_DEFAULT                                  0x00000000
#define regVPCM_GAMCOR_RAMA_REGION_32_33_DEFAULT                                  0x00000000
#define regVPCM_HDR_MULT_COEF_DEFAULT                                             0x0001f000
#define regVPCM_MEM_PWR_CTRL_DEFAULT                                              0x00000000
#define regVPCM_MEM_PWR_STATUS_DEFAULT                                            0x00000000
#define regVPCM_DEALPHA_DEFAULT                                                   0x00000000
#define regVPCM_COEF_FORMAT_DEFAULT                                               0x00000000
#define regVPCM_TEST_DEBUG_INDEX_DEFAULT                                          0x00000000
#define regVPCM_TEST_DEBUG_DATA_DEFAULT                                           0x00000000


// addressBlock: vpe_vpep_vpdpp0_dispdec_vpdpp_top_dispdec
#define regVPDPP_CONTROL_DEFAULT                                                  0x70000000
#define regVPDPP_SOFT_RESET_DEFAULT                                               0x00000000
#define regVPDPP_CRC_VAL_R_G_DEFAULT                                              0x00000000
#define regVPDPP_CRC_VAL_B_A_DEFAULT                                              0x00000000
#define regVPDPP_CRC_CTRL_DEFAULT                                                 0x00000000
#define regVPHOST_READ_CONTROL_DEFAULT                                            0x00000000
#define regVPDPP_DEBUG_SEL_DEFAULT                                                0x00000000
#define regVPDPP_DEBUG_SPARE_DEFAULT                                              0x00000000
#define regVPDPP_TEST_DEBUG_INDEX_DEFAULT                                         0x00000000
#define regVPDPP_TEST_DEBUG_DATA_DEFAULT                                          0x00000000


// addressBlock: vpe_vpep_vpmpc_vpmpcc0_dispdec
#define regVPMPCC_TOP_SEL_DEFAULT                                                 0x0000000f
#define regVPMPCC_BOT_SEL_DEFAULT                                                 0x0000000f
#define regVPMPCC_VPOPP_ID_DEFAULT                                                0x0000000f
#define regVPMPCC_CONTROL_DEFAULT                                                 0xffff0461
#define regVPMPCC_TOP_GAIN_DEFAULT                                                0x0001f000
#define regVPMPCC_BOT_GAIN_INSIDE_DEFAULT                                         0x0001f000
#define regVPMPCC_BOT_GAIN_OUTSIDE_DEFAULT                                        0x0001f000
#define regVPMPCC_MOVABLE_CM_LOCATION_CONTROL_DEFAULT                             0x00000001
#define regVPMPCC_BG_R_CR_DEFAULT                                                 0x00000000
#define regVPMPCC_BG_G_Y_DEFAULT                                                  0x00000000
#define regVPMPCC_BG_B_CB_DEFAULT                                                 0x00000000
#define regVPMPCC_MEM_PWR_CTRL_DEFAULT                                            0x00000010
#define regVPMPCC_STATUS_DEFAULT                                                  0x00000000
#define regVPMPCC_TEST_DEBUG_INDEX_DEFAULT                                        0x00000000
#define regVPMPCC_TEST_DEBUG_DATA_DEFAULT                                         0x00000000


// addressBlock: vpe_vpep_vpmpc_vpmpc_cfg_dispdec
#define regVPMPC_CLOCK_CONTROL_DEFAULT                                            0x00000000
#define regVPMPC_SOFT_RESET_DEFAULT                                               0x00000000
#define regVPMPC_CRC_CTRL_DEFAULT                                                 0x00000000
#define regVPMPC_CRC_SEL_CONTROL_DEFAULT                                          0x00000000
#define regVPMPC_CRC_RESULT_AR_DEFAULT                                            0x00000000
#define regVPMPC_CRC_RESULT_GB_DEFAULT                                            0x00000000
#define regVPMPC_CRC_RESULT_C_DEFAULT                                             0x00000000
#define regVPMPC_DEBUG_CONTROL_DEFAULT                                            0x0000ff00
#define regVPMPCC_DEBUG_DATA_SELECT_DEFAULT                                       0x3f2f1f0f
#define regVPMPC_BYPASS_BG_AR_DEFAULT                                             0x00000000
#define regVPMPC_BYPASS_BG_GB_DEFAULT                                             0x00000000
#define regVPMPC_HOST_READ_CONTROL_DEFAULT                                        0x00000000
#define regVPMPC_PENDING_STATUS_MISC_DEFAULT                                      0x00000000
#define regVPMPC_CFG_TEST_DEBUG_INDEX_DEFAULT                                     0x00000000
#define regVPMPC_CFG_TEST_DEBUG_DATA_DEFAULT                                      0x00000000


// addressBlock: vpe_vpep_vpmpc_vpmpcc_ogam0_dispdec
#define regVPMPCC_OGAM_CONTROL_DEFAULT                                            0x00000000
#define regVPMPCC_OGAM_LUT_INDEX_DEFAULT                                          0x00000000
#define regVPMPCC_OGAM_LUT_DATA_DEFAULT                                           0x00000000
#define regVPMPCC_OGAM_LUT_CONTROL_DEFAULT                                        0x00000007
#define regVPMPCC_OGAM_RAMA_START_CNTL_B_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_START_CNTL_G_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_START_CNTL_R_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_START_SLOPE_CNTL_B_DEFAULT                            0x00000000
#define regVPMPCC_OGAM_RAMA_START_SLOPE_CNTL_G_DEFAULT                            0x00000000
#define regVPMPCC_OGAM_RAMA_START_SLOPE_CNTL_R_DEFAULT                            0x00000000
#define regVPMPCC_OGAM_RAMA_START_BASE_CNTL_B_DEFAULT                             0x00000000
#define regVPMPCC_OGAM_RAMA_START_BASE_CNTL_G_DEFAULT                             0x00000000
#define regVPMPCC_OGAM_RAMA_START_BASE_CNTL_R_DEFAULT                             0x00000000
#define regVPMPCC_OGAM_RAMA_END_CNTL1_B_DEFAULT                                   0x00000000
#define regVPMPCC_OGAM_RAMA_END_CNTL2_B_DEFAULT                                   0x00000000
#define regVPMPCC_OGAM_RAMA_END_CNTL1_G_DEFAULT                                   0x00000000
#define regVPMPCC_OGAM_RAMA_END_CNTL2_G_DEFAULT                                   0x00000000
#define regVPMPCC_OGAM_RAMA_END_CNTL1_R_DEFAULT                                   0x00000000
#define regVPMPCC_OGAM_RAMA_END_CNTL2_R_DEFAULT                                   0x00000000
#define regVPMPCC_OGAM_RAMA_OFFSET_B_DEFAULT                                      0x00000000
#define regVPMPCC_OGAM_RAMA_OFFSET_G_DEFAULT                                      0x00000000
#define regVPMPCC_OGAM_RAMA_OFFSET_R_DEFAULT                                      0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_0_1_DEFAULT                                    0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_2_3_DEFAULT                                    0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_4_5_DEFAULT                                    0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_6_7_DEFAULT                                    0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_8_9_DEFAULT                                    0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_10_11_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_12_13_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_14_15_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_16_17_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_18_19_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_20_21_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_22_23_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_24_25_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_26_27_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_28_29_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_30_31_DEFAULT                                  0x00000000
#define regVPMPCC_OGAM_RAMA_REGION_32_33_DEFAULT                                  0x00000000
#define regVPMPCC_GAMUT_REMAP_COEF_FORMAT_DEFAULT                                 0x00000000
#define regVPMPCC_GAMUT_REMAP_MODE_DEFAULT                                        0x00000000
#define regVPMPC_GAMUT_REMAP_C11_C12_A_DEFAULT                                    0x00000000
#define regVPMPC_GAMUT_REMAP_C13_C14_A_DEFAULT                                    0x00000000
#define regVPMPC_GAMUT_REMAP_C21_C22_A_DEFAULT                                    0x00000000
#define regVPMPC_GAMUT_REMAP_C23_C24_A_DEFAULT                                    0x00000000
#define regVPMPC_GAMUT_REMAP_C31_C32_A_DEFAULT                                    0x00000000
#define regVPMPC_GAMUT_REMAP_C33_C34_A_DEFAULT                                    0x00000000
#define regVPMPCC_OGAM_TEST_DEBUG_INDEX_DEFAULT                                   0x00000000
#define regVPMPCC_OGAM_TEST_DEBUG_DATA_DEFAULT                                    0x00000000


// addressBlock: vpe_vpep_vpmpc_vpmpcc_mcm0_dispdec
#define regVPMPCC_MCM_SHAPER_CONTROL_DEFAULT                                      0x00000000
#define regVPMPCC_MCM_SHAPER_OFFSET_R_DEFAULT                                     0x00000000
#define regVPMPCC_MCM_SHAPER_OFFSET_G_DEFAULT                                     0x00000000
#define regVPMPCC_MCM_SHAPER_OFFSET_B_DEFAULT                                     0x00000000
#define regVPMPCC_MCM_SHAPER_SCALE_R_DEFAULT                                      0x00007000
#define regVPMPCC_MCM_SHAPER_SCALE_G_B_DEFAULT                                    0x70007000
#define regVPMPCC_MCM_SHAPER_LUT_INDEX_DEFAULT                                    0x00000000
#define regVPMPCC_MCM_SHAPER_LUT_DATA_DEFAULT                                     0x00000000
#define regVPMPCC_MCM_SHAPER_LUT_WRITE_EN_MASK_DEFAULT                            0x00000007
#define regVPMPCC_MCM_SHAPER_RAMA_START_CNTL_B_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_START_CNTL_G_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_START_CNTL_R_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_END_CNTL_B_DEFAULT                              0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_END_CNTL_G_DEFAULT                              0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_END_CNTL_R_DEFAULT                              0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_0_1_DEFAULT                              0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_2_3_DEFAULT                              0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_4_5_DEFAULT                              0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_6_7_DEFAULT                              0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_8_9_DEFAULT                              0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_10_11_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_12_13_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_14_15_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_16_17_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_18_19_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_20_21_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_22_23_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_24_25_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_26_27_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_28_29_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_30_31_DEFAULT                            0x00000000
#define regVPMPCC_MCM_SHAPER_RAMA_REGION_32_33_DEFAULT                            0x00000000
#define regVPMPCC_MCM_3DLUT_MODE_DEFAULT                                          0x00000000
#define regVPMPCC_MCM_3DLUT_INDEX_DEFAULT                                         0x00000000
#define regVPMPCC_MCM_3DLUT_DATA_DEFAULT                                          0x00000000
#define regVPMPCC_MCM_3DLUT_DATA_30BIT_DEFAULT                                    0x00000000
#define regVPMPCC_MCM_3DLUT_READ_WRITE_CONTROL_DEFAULT                            0x0000000f
#define regVPMPCC_MCM_3DLUT_OUT_NORM_FACTOR_DEFAULT                               0x00008008
#define regVPMPCC_MCM_3DLUT_OUT_OFFSET_R_DEFAULT                                  0x3c000000
#define regVPMPCC_MCM_3DLUT_OUT_OFFSET_G_DEFAULT                                  0x3c000000
#define regVPMPCC_MCM_3DLUT_OUT_OFFSET_B_DEFAULT                                  0x3c000000
#define regVPMPCC_MCM_1DLUT_CONTROL_DEFAULT                                       0x00000000
#define regVPMPCC_MCM_1DLUT_LUT_INDEX_DEFAULT                                     0x00000000
#define regVPMPCC_MCM_1DLUT_LUT_DATA_DEFAULT                                      0x00000000
#define regVPMPCC_MCM_1DLUT_LUT_CONTROL_DEFAULT                                   0x00000007
#define regVPMPCC_MCM_1DLUT_RAMA_START_CNTL_B_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_START_CNTL_G_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_START_CNTL_R_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_START_SLOPE_CNTL_B_DEFAULT                       0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_START_SLOPE_CNTL_G_DEFAULT                       0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_START_SLOPE_CNTL_R_DEFAULT                       0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_START_BASE_CNTL_B_DEFAULT                        0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_START_BASE_CNTL_G_DEFAULT                        0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_START_BASE_CNTL_R_DEFAULT                        0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_END_CNTL1_B_DEFAULT                              0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_END_CNTL2_B_DEFAULT                              0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_END_CNTL1_G_DEFAULT                              0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_END_CNTL2_G_DEFAULT                              0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_END_CNTL1_R_DEFAULT                              0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_END_CNTL2_R_DEFAULT                              0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_OFFSET_B_DEFAULT                                 0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_OFFSET_G_DEFAULT                                 0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_OFFSET_R_DEFAULT                                 0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_0_1_DEFAULT                               0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_2_3_DEFAULT                               0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_4_5_DEFAULT                               0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_6_7_DEFAULT                               0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_8_9_DEFAULT                               0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_10_11_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_12_13_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_14_15_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_16_17_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_18_19_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_20_21_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_22_23_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_24_25_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_26_27_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_28_29_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_30_31_DEFAULT                             0x00000000
#define regVPMPCC_MCM_1DLUT_RAMA_REGION_32_33_DEFAULT                             0x00000000
#define regVPMPCC_MCM_MEM_PWR_CTRL_DEFAULT                                        0x00101010
#define regVPMPCC_MCM_TEST_DEBUG_INDEX_DEFAULT                                    0x00000000
#define regVPMPCC_MCM_TEST_DEBUG_DATA_DEFAULT                                     0x00000000


// addressBlock: vpe_vpep_vpmpc_vpmpc_ocsc_dispdec
#define regVPMPC_OUT0_MUX_DEFAULT                                                 0x0000000f
#define regVPMPC_OUT0_FLOAT_CONTROL_DEFAULT                                       0x00000000
#define regVPMPC_OUT0_DENORM_CONTROL_DEFAULT                                      0x00fff000
#define regVPMPC_OUT0_DENORM_CLAMP_G_Y_DEFAULT                                    0x00fff000
#define regVPMPC_OUT0_DENORM_CLAMP_B_CB_DEFAULT                                   0x00fff000
#define regVPMPC_OUT_CSC_COEF_FORMAT_DEFAULT                                      0x00000000
#define regVPMPC_OUT0_CSC_MODE_DEFAULT                                            0x00000000
#define regVPMPC_OUT0_CSC_C11_C12_A_DEFAULT                                       0x00000000
#define regVPMPC_OUT0_CSC_C13_C14_A_DEFAULT                                       0x00000000
#define regVPMPC_OUT0_CSC_C21_C22_A_DEFAULT                                       0x00000000
#define regVPMPC_OUT0_CSC_C23_C24_A_DEFAULT                                       0x00000000
#define regVPMPC_OUT0_CSC_C31_C32_A_DEFAULT                                       0x00000000
#define regVPMPC_OUT0_CSC_C33_C34_A_DEFAULT                                       0x00000000
#define regVPMPC_OCSC_TEST_DEBUG_INDEX_DEFAULT                                    0x00000000
#define regVPMPC_OCSC_TEST_DEBUG_DATA_DEFAULT                                     0x00000000


// addressBlock: vpe_vpep_vpopp_vpfmt0_dispdec
#define regVPFMT_CLAMP_COMPONENT_R_DEFAULT                                        0x00000000
#define regVPFMT_CLAMP_COMPONENT_G_DEFAULT                                        0x00000000
#define regVPFMT_CLAMP_COMPONENT_B_DEFAULT                                        0x00000000
#define regVPFMT_DYNAMIC_EXP_CNTL_DEFAULT                                         0x00000000
#define regVPFMT_CONTROL_DEFAULT                                                  0x00000000
#define regVPFMT_BIT_DEPTH_CONTROL_DEFAULT                                        0x00000000
#define regVPFMT_DITHER_RAND_R_SEED_DEFAULT                                       0x00000000
#define regVPFMT_DITHER_RAND_G_SEED_DEFAULT                                       0x00000099
#define regVPFMT_DITHER_RAND_B_SEED_DEFAULT                                       0x000000dd
#define regVPFMT_CLAMP_CNTL_DEFAULT                                               0x00000000
#define regVPFMT_DEBUG_CNTL_DEFAULT                                               0x00000000
#define regVPFMT_TEST_DEBUG_INDEX_DEFAULT                                         0x00000000
#define regVPFMT_TEST_DEBUG_DATA_DEFAULT                                          0x00000000


// addressBlock: vpe_vpep_vpopp_vpopp_pipe0_dispdec
#define regVPOPP_PIPE_CONTROL_DEFAULT                                             0x00000000
#define regVPOPP_PIPE_SPARE_DEBUG_DEFAULT                                         0x00000000
#define regVPOPP_PIPE_TEST_DEBUG_INDEX_DEFAULT                                    0x00000000
#define regVPOPP_PIPE_TEST_DEBUG_DATA_DEFAULT                                     0x00000000


// addressBlock: vpe_vpep_vpopp_vpopp_pipe_crc0_dispdec
#define regVPOPP_PIPE_CRC_CONTROL_DEFAULT                                         0x00000000
#define regVPOPP_PIPE_CRC_MASK_DEFAULT                                            0x0000ffff
#define regVPOPP_PIPE_CRC_RESULT0_DEFAULT                                         0x00000000
#define regVPOPP_PIPE_CRC_RESULT1_DEFAULT                                         0x00000000
#define regVPOPP_PIPE_CRC_RESULT2_DEFAULT                                         0x00000000


// addressBlock: vpe_vpep_vpopp_vpopp_top_dispdec
#define regVPOPP_TOP_CLK_CONTROL_DEFAULT                                          0x00000000
#define regVPOPP_DEBUG_CONTROL_DEFAULT                                            0x00000000
#define regVPOPP_TOP_SPARE_DEBUG_DEFAULT                                          0x00000000
#define regVPOPP_TOP_TEST_DEBUG_INDEX_DEFAULT                                     0x00000000
#define regVPOPP_TOP_TEST_DEBUG_DATA_DEFAULT                                      0x00000000


// addressBlock: vpe_vpep_vpcdc_cdc_dispdec
#define regVPEP_MGCG_CNTL_DEFAULT                                                 0x00000000
#define regVPCDC_SOFT_RESET_DEFAULT                                               0x00000000
#define regVPCDC_FE0_SURFACE_CONFIG_DEFAULT                                       0x00000000
#define regVPCDC_FE0_CROSSBAR_CONFIG_DEFAULT                                      0x00000000
#define regVPCDC_FE0_VIEWPORT_START_CONFIG_DEFAULT                                0x00000000
#define regVPCDC_FE0_VIEWPORT_DIMENSION_CONFIG_DEFAULT                            0x00000000
#define regVPCDC_FE0_VIEWPORT_START_C_CONFIG_DEFAULT                              0x00000000
#define regVPCDC_FE0_VIEWPORT_DIMENSION_C_CONFIG_DEFAULT                          0x00000000
#define regVPCDC_BE0_P2B_CONFIG_DEFAULT                                           0x00000036
#define regVPCDC_BE0_GLOBAL_SYNC_CONFIG_DEFAULT                                   0x00000000
#define regVPCDC_GLOBAL_SYNC_TRIGGER_DEFAULT                                      0x00000000
#define regVPCDC_VREADY_STATUS_DEFAULT                                            0x00000000
#define regVPEP_MEM_GLOBAL_PWR_REQ_CNTL_DEFAULT                                   0x00000000
#define regVPFE_MEM_PWR_CNTL_DEFAULT                                              0x00000000
#define regVPBE_MEM_PWR_CNTL_DEFAULT                                              0x00000000
#define regVPEP_RBBMIF_TIMEOUT_DEFAULT                                            0x01f00100
#define regVPEP_RBBMIF_STATUS_DEFAULT                                             0x00000000
#define regVPEP_RBBMIF_TIMEOUT_DIS_DEFAULT                                        0x00000000
#define regVPCDC_DEBUG_CTRL0_DEFAULT                                              0x00000000
#define regVPCDC_DEBUG_CTRL1_DEFAULT                                              0x00000000
#define regVPCDC_TEST_DEBUG_INDEX_DEFAULT                                         0x00000000
#define regVPCDC_TEST_DEBUG_DATA_DEFAULT                                          0x00000000


// addressBlock: vpe_vpep_vpcdc_vpcdc_dcperfmon_dc_perfmon_dispdec
#define regPERFCOUNTER_CNTL_DEFAULT                                               0x00000000
#define regPERFCOUNTER_CNTL2_DEFAULT                                              0x00000000
#define regPERFCOUNTER_STATE_DEFAULT                                              0x00000000
#define regPERFMON_CNTL_DEFAULT                                                   0x00000100
#define regPERFMON_CNTL2_DEFAULT                                                  0x00000000
#define regPERFMON_CVALUE_INT_MISC_DEFAULT                                        0x00000000
#define regPERFMON_CVALUE_LOW_DEFAULT                                             0x00000000
#define regPERFMON_HI_DEFAULT                                                     0x00000000
#define regPERFMON_LOW_DEFAULT                                                    0x00000000
#define regPERFMON_TEST_DEBUG_INDEX_DEFAULT                                       0x00000000
#define regPERFMON_TEST_DEBUG_DATA_DEFAULT                                        0x00000000


// addressBlock: dc_perfmon_dc_perfmondebugind
#define ixPERFMON_DEBUG_ID_DEFAULT                                               0x00000000
#define ixPERFMON_DEBUG01_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG02_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG03_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG04_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG05_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG06_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG07_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG08_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG09_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG0A_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG0B_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG0C_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG0D_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG0E_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG0F_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG10_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG11_DEFAULT                                                0x00000000
#define ixPERFMON_DEBUG12_DEFAULT                                                0x00000000


// addressBlock: vpfmt0_vpfmtdebugind
#define ixVPFMT_DEBUG_ID_DEFAULT                                                 0x00000000
#define ixVPFMT_DEBUG0_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG1_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG2_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG3_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG4_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG5_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG6_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG7_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG8_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG9_DEFAULT                                                   0x00000000
#define ixVPFMT_DEBUG10_DEFAULT                                                  0x00000000
#define ixVPFMT_DEBUG11_DEFAULT                                                  0x00000000


// addressBlock: vpopp_pipe0_vpopppipedebugind
#define ixVPOPP_PIPE_DEBUG_ID_DEFAULT                                            0x00000000
#define ixVPOPP_PIPE_DEBUG_0_DEFAULT                                             0x00000000
#define ixVPOPP_PIPE_DEBUG_1_DEFAULT                                             0x00000000
#define ixVPOPP_PIPE_DEBUG_2_DEFAULT                                             0x00000000


// addressBlock: vpopp_top_vpopp_topdebugind
#define ixVPOPP_TOP_DEBUG_ID_DEFAULT                                             0x00000000

#endif
