dp2(8)          g7<1>.xF        g7<4>.xyyyF     g7<4>.xyyyF     { align16 };
dp2(8)          m5<1>.xF        g1<0>.yF        g1<0>.yF        { align16 };
dp2(8)          m5<1>.yzF       g1<0>.xF        g1<0>.zwwwF     { align16 NoDDClr };
dp2(8)          m5<1>.wF        g1<0>.ywwwF     g1<0>.wyyyF     { align16 NoDDChk };
dp2(8)          g4<1>.yF        g1<0>.xyyyF     g1.4<0>.xyyyF   { align16 NoDDClr };
dp2(8)          g4<1>.zF        g1<0>.xyyyF     g1.4<0>.zwwwF   { align16 NoDDClr,NoDDChk };
dp2(8)          g4<1>.wF        g1<0>.xyyyF     g2<0>.xyyyF     { align16 NoDDChk };
