/*
 * Copyright © 2024 Intel Corporation
 * SPDX-License-Identifier: MIT
 */

#ifndef EXECUTOR_H
#error This file must be included via executor.h
#endif

void genX(emit_execute)(executor_context *ec, const executor_params *params);
