dph(8)          m5<1>.xF        g4<4>.xyzxF     g5<4>F          { align16 };
dph.sat(8)      m5<1>F          g1<0>.xyzxF     g3<4>F          { align16 };
dph(8)          g5<1>.xF        g4<4>.xyzxF     g1<0>F          { align16 NoDDClr };
dph(8)          g5<1>.yF        g4<4>.xyzxF     g1.4<0>F        { align16 NoDDClr,NoDDChk };
dph(8)          g6<1>.wF        g5<4>.xyzxF     g2.4<0>F        { align16 NoDDChk };
