/*
 * Copyright © 2016 Rob Clark <robclark@freedesktop.org>
 * Copyright © 2018 Google, Inc.
 * SPDX-License-Identifier: MIT
 *
 * Authors:
 *    Rob Clark <robclark@freedesktop.org>
 */

#ifndef FD6_SCREEN_H_
#define FD6_SCREEN_H_

#include "freedreno_screen.h"

EXTERNC void fd6_screen_init(struct pipe_screen *pscreen);

#endif /* FD6_SCREEN_H_ */
